/*
 * rmessaged
 *
 * A daemon that sends periodic signals to processes who request such
 * a service. 
 * by Dave Lowell
 */

/**
 ** Includes
 **/

#include "rmessage.h"

/**
 ** Types
 **/

typedef struct proc_s {
	int	pid;
	struct proc_s *next;
} proc;

/**
 ** Globals
 **/

proc	*list = NULL;

/**
 ** Functions
 **/

void* alloc(int bytes)
{
	void*	mem;

	mem = malloc(bytes);
	if (mem == NULL) {
		fprintf(stderr, "rmessaged: out of memory\n");
		exit(1);
	}
	return mem;
}

void add_proc(int pid)
{
	proc	*p;

	/* check if this proc is already in list */
	for (p=list; p != NULL; p = p->next)
		if (p->pid == pid) return;

	/* add it */
	p = alloc(sizeof(proc));
	p->pid = pid;
	p->next = list;
	list = p;
	fprintf(stderr, "rmessaged: added proc %d\n", pid);
}

void delete_proc(int pid)
{
	proc	*p, *prev;

	for (p=list, prev=NULL; p != NULL; p = p->next) {
		if (p->pid == pid) {
			if (prev == NULL) 
				list = p->next;
			else 
				prev->next = p->next;
			free(p);
			return;
		}
		prev = p;
	}
}

void signal_all(void)
{
	int	res;
	proc	*p, *n;

	/* send signal to every process on list */
	for (p=list; p != NULL; p = n) {
		n = p->next;
		res = kill(p->pid, RM_TIMER_SIG);
		if (res < 0 && errno == ESRCH) {
			/* process has exited */
			delete_proc(p->pid);
			fprintf(stderr, "rmessaged: process %d done\n", p->pid);
		}
		else if (res < 0)
			perror("rmessaged: kill()");
	}
}

void print_list(void)
{
	proc	*p;

	fprintf(stderr, "list -");
	for (p=list; p != NULL; p = p->next)
		fprintf(stderr, "> %d -", p->pid);
	fprintf(stderr, "v\n");
}


void main(void)
{
	int	sock, res;
	fd_set	f;
	struct timeval tout = {0, RM_TIMER_PERIOD*1000};
	struct timeval poll = {0, 0};
	struct sockaddr_in addr;

	fprintf(stderr, "rmessaged: initializing\n");

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("rmessaged: socket()");
		exit(1);
	}

	bzero(&addr, sizeof(addr));
	addr.sin_family 	= AF_INET;
	addr.sin_addr.s_addr 	= inet_addr("127.0.0.1");
	addr.sin_port 		= htons(RM_DAEMON_PORT);

	res = bind(sock, (struct sockaddr*) &addr, sizeof(addr));
	if (res < 0) {
		perror("rmessaged: bind()");
		exit(1);
	}

	fprintf(stderr, "rmessaged: ready\n");

	for (;;) {

		/* Block for RM_TIMER_PERIOD ms */
		res = select(0, NULL, NULL, NULL, &tout);
		if (res < 0) {
			perror("rmessaged: select()");
			continue;
		}

		/* Send RM_TIMER_SIG signal to all current procs */
		signal_all();

		/* Check for new procs to add */
		FD_ZERO(&f);
		FD_SET(sock, &f);
		while (select(OPEN_MAX, &f, NULL, NULL, &poll) > 0) {
			int	pid, len;

			len = sizeof(addr);
			res = recvfrom(sock, &pid, sizeof(pid), 0, 
				       (struct sockaddr*) &addr, &len);
			if (res < 0) {
				perror("rmessaged: recfrom()");
				exit(1);
			}

			add_proc(pid);

			res = sendto(sock, "", 1, 0, (struct sockaddr*) &addr,
				     sizeof(addr));
			if (res < 0) {
				perror("rmessaged: sendto()");
				exit(1);
			}
		}
	}
}


