/*
 * Rmessage library. By Dave Lowell.
 *
 * This library augments traditional UDP by providing reliable, ordered,
 * delivery of datagram messages. All acknowledgements are done 
 * asynchronously, and an attempt is made to piggyback acks on return 
 * messages as appropriate. 
 */

/**
 ** Includes
 **/

#include <sys/syscall.h>
#include <sys/time.h>
#include <signal.h>
#include "rmessage.h"
#include "hash.h"


/**
 ** Global Variables
 **/

hash_table	*connections;
hash_table	*seqno_table;
hash_table	*addr_table;
outstanding_msg	*outs_h;
outstanding_msg	*outs_t;
connection_info	*to_ack[RM_MAX_DESCRIPTORS];
int		curr_ack_index;
int		descriptors[RM_MAX_DESCRIPTORS+1];
int		sock_retrans[RM_MAX_DESCRIPTORS];
int		msgs_on_sock[RM_MAX_DESCRIPTORS];
int		bytes_on_sock[RM_MAX_DESCRIPTORS];
int		local_ports[RM_MAX_DESCRIPTORS];
sockbuf		sock_queues[RM_MAX_DESCRIPTORS];
int		is_rm_sock[RM_MAX_DESCRIPTORS];
int		bytes_buffered[RM_MAX_DESCRIPTORS];
int		num_desc = 0;
int		rm_signal_flag = FALSE;

tracent		trace[100000];
int		trace_i = 0;

char 		*sfile = NULL;
int 		sline = 0;


/**
 ** Support Functions
 **/

int select(int maxfdp1, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
		struct timeval *tvptr)
{
	if(readfds && !FD_ISSET(maxfdp1+2, readfds))
	{
		/*fprintf(stderr, "got select() from user routine\n"); */
		return mt_select(maxfdp1, readfds, writefds, exceptfds,
					tvptr);
	}
	else
	{
		/*fprintf(stderr, "got select() from rm, or null readfds\n");
		*/
		return syscall(SYS_select, maxfdp1, readfds, writefds, 
				exceptfds, tvptr);
	}
}

/* returns current time in milliseconds */
long get_time(void)
{
	struct timeval t;
	long s, u;

	if (gettimeofday(&t, NULL) < 0)
		perror("gettimeofday");
	s = t.tv_sec;
	u = t.tv_usec;
	return((s * 1000L) + (u / 1000L));
}

/* returns number of seconds between two times, each expressed in millis */
double elapsed_time(long start, long finish)
{
	long    secs;
	long    millis;

	secs = (finish - start) / 1000L;
	millis = (finish - start) % 1000L;
	return ((double)secs) + (((double)millis) * 0.001);
}



void* alloc(int size)
{
	void	*p;

	p = malloc(size);
	if (p == NULL) {
		fprintf(stderr, "rmessage: out of memory.\n");
		exit(1);
	}
	return p;
}

void rm_lock(void)
{
	int	res;
	sigset_t m;

	sigemptyset(&m);
	sigaddset(&m, RM_TIMER_SIG);
	res = sigprocmask(SIG_BLOCK, &m, NULL);
	if (res < 0) perror("rm_lock");
}

void rm_unlock(void)
{
	int	res;
	sigset_t m;

	sigemptyset(&m);
	sigaddset(&m, RM_TIMER_SIG);
	res = sigprocmask(SIG_UNBLOCK, &m, NULL);
	if (res < 0) perror("rm_lock");
}

/*
 * get_address() converts its host argument string (either a hostname or an
 * IP address in dotted decimal format), into an internet address and returns
 * it, or FALSE on failure.
 */     
unsigned long get_address(char* host)
{
	struct in_addr  *addr_ptr; 
	unsigned long   inaddr;

	/*              
	 * First check if host string we've been passed is in
	 * dotted decimal format. Failing that, we'll get the 
	 * IP from the system.
	 */     
	if ((int)(inaddr = inet_addr(host)) == INADDR_NONE) {
		struct hostent  *hp;
		struct in_addr	*addr;

		hp = gethostbyname(host);
		if (hp == NULL) {
			return FALSE;
		}
		addr = (struct in_addr*) hp->h_addr;
		return addr->s_addr;
	}
	else {
		return inaddr;
	}
}


int get_local_port(int sock)
{
	int	port, res, len;
	struct sockaddr_in name;

	port = local_ports[sock];
	if (port == 0) {
		len = sizeof(name);
		res = getsockname(sock, (struct sockaddr*) &name, &len);
		if (res < 0) {
			perror("getsockname");
			return -1;
		}
		port = ntohs(name.sin_port);
		if (port == 0) {
			/*
			 * This socket has not yet been bound to a port.
			 * We'll just go ahead and bind it to a system
			 * assigned port, and then request the port num from
			 * the system.
			 */
			res = rm_bind(sock, 0);
			if (res < 0) return -1;
			len = sizeof(name);
			res = getsockname(sock, (struct sockaddr*) &name, &len);
			if (res < 0) {
				perror("getsockname");
				return -1;
			}
			port = ntohs(name.sin_port);
		}
		local_ports[sock] = port;
	}
	return port;
}


connection_info* get_conn_info(int sock, u_long ip, int port)
{
	int		loc_port;
	char		key[32];
	connection_info *c;

	loc_port = get_local_port(sock);
	sprintf(key, "%d:%lu:%d", loc_port, ip, port);
	c = hash_find(connections, key);
	if (c == NULL) {
		c = alloc(sizeof(connection_info));
		c->received 	= RM_MAX_ID;
		c->mcount 	= 0;
		c->do_ack 	= FALSE;
		c->ack_index 	= -1;
		c->nack_seq 	= RM_MAX_ID;
		c->closing 	= FALSE;
		c->oo_table 	= alloc(sizeof(hash_table));
		hash_init(c->oo_table, HASH_ONE_WORD_KEYS);
		hash_insert(connections, key, c);
	}
	return c;
}


u_int new_seqno(int sock, u_long ip, int port)
{
	char	key[32];
	int	*p, loc_port;

	loc_port = get_local_port(sock);
	sprintf(key, "%d:%lu:%d", loc_port, ip, port);
	p = hash_find(seqno_table, key);
	if (p == NULL) {
		p = alloc(sizeof(int));
		*p = 0;
		hash_insert(seqno_table, key, p);
	}
	return (*p)++;
}


rm_msg* build_msg(int sock, char *data, int len, u_long ip, int port, int flags)
{
	u_int		seqno;
	rm_msg		*m;
	connection_info *c;

	seqno = new_seqno(sock, ip, port);
	c = get_conn_info(sock, ip, port);

	/* 
	 * Since this message will carry piggybacked ack,
	 * disable any pending asychronous or flow ack to the 
	 * same host+port from this local port.
	 */
	c->do_ack = FALSE; 
	c->mcount = 0; 

	/* Create message and fill in fields */
	m = alloc(len + RM_HEADER_SIZE);
	m->id 		= seqno;
	m->type		= (flags & MSG_EOF) ? RM_EOF : RM_REG;
	m->len 		= len;
	m->rcvd 	= c->received; /* ack all msgs up to c->received */
	m->retrans 	= FALSE;
	bcopy(data, m->data, len);

	return m;
}


void add_msg_to_tables(rm_msg* m, int s, u_long ip, int port, int flags)
{
	outstanding_msg	*o;

	o = alloc(sizeof(outstanding_msg));

	o->m 		= m;
	o->sock 	= s;
	o->flags 	= flags;
	o->ip 		= ip;
	o->port 	= port;
	o->ticks 	= 0;
	o->next 	= NULL;
	o->prev 	= NULL;

	if (outs_h == NULL) {
		outs_h = o;
		outs_t = o;
	}
	else {
		o->prev = outs_t;
		outs_t->next = o;
		outs_t = o;
	}
}


/*
 * Receive an incoming message into two buffers without extra copying.
 */
int recv_vec(int sock, void* b1, int b1_len, void* b2, int b2_len, 
	     int flags, struct sockaddr_in* addr)
{
	struct msghdr	h;
	struct iovec	sg[2];
	int		val;

	sg[0].iov_base = b1;
	sg[0].iov_len = b1_len;
	sg[1].iov_base = b2;
	sg[1].iov_len = b2_len;

	h.msg_name = (caddr_t) addr;
	h.msg_namelen = (addr==NULL?0:sizeof(struct sockaddr_in));
	h.msg_iov = sg;
	h.msg_iovlen = 2;
	h.msg_control = NULL;
	h.msg_controllen = 0;
	h.msg_flags = 0;

	/*
	 * Retry this receive if it was interrupted by a signal
	 */
	while ((val = recvmsg(sock, &h, flags)) < 0 && errno == EINTR);
	return val;
}


int udp_recv(int sock, void* b, int len, int flags, struct sockaddr_in* addr)
{
	int		val, nlen;

	nlen = (addr == NULL ? 0 : sizeof(struct sockaddr_in));

	/*
	 * Retry this receive if it was interrupted by a signal
	 */
	while ((val = recvfrom(sock, b, len, flags, (struct sockaddr*)addr, 
	       &nlen)) < 0 && errno == EINTR);
	return val;
}


int udp_send(int sock, void* data, int len, int flags, u_long ip, int port)
{
	int		res;
	struct timeval	t = {0, 1}; /* 1 us */
	struct sockaddr_in d;

	/* Fill in destination sockaddr struct */
	d.sin_family 		= AF_INET;
	d.sin_port 		= htons(port);
	d.sin_addr.s_addr 	= ip;

	/*
	 * The Freebsd messaging code easily fills up its available
	 * message buffers (~50 of them). If there are no available buffers
	 * we'll sleep for 1 clock tick and then try again.
	 */
	while ((res = sendto(sock, data, len, flags, (struct sockaddr*) &d, 
		sizeof(d))) < 0 && errno == ENOBUFS) {
		select(0, NULL, NULL, NULL, &t);
	}
	return res;
}


/* 
 * Is message 'id' within acked window? To be in the window, the message
 * must be between 'upto' and 'upto' - RM_WINDOW, even if that window wraps
 * around to higher ids.
 */
int in_wind(u_int id, u_int upto)
{
	u_int	lim;

	if (upto <= RM_WINDOW) {
		lim = RM_MAX_ID - (RM_WINDOW - upto);
		if (id <= upto || id >= lim) 
			return TRUE;
		else 
			return FALSE;
	}
	else {
		lim = upto - RM_WINDOW;
		if (id <= upto && id >= lim) 
			return TRUE;
		else 
			return FALSE;
	}
}


/*
 * Is message 'x' ahead of message 'y' in sequence?
 */
int is_ahead(u_int x, u_int y)
{
	if (x == y)
		return FALSE;
	else
		return in_wind(y, x);
}


/*
 * process_acks() removes messages from retransmit list that have
 * been acknowledged. Signals should be off (rm_lock held) during this
 * call.
 */
void process_acks(u_long ip, int port, u_int upto)
{
	outstanding_msg *o, *n;

	for (o = outs_h; o != NULL; o = n) {
		n = o->next;
		if (o->ip == ip && o->port == port && in_wind(o->m->id, upto)) {
			/*
			 * Decrement outstanding msgs/sock and 
			 * bytes/sock counters.
			 */
			msgs_on_sock[o->sock]--; 
			bytes_on_sock[o->sock] -= o->m->len + RM_HEADER_SIZE; 
			free(o->m); /* we're done with message */
			if (outs_h == o) {
				if (o->next != NULL)
					o->next->prev = NULL;
				outs_h = o->next;
				if (outs_h == NULL)
					outs_t = NULL;
			}
			else if (outs_t == o) {
				if (o->prev != NULL)
					o->prev->next = NULL;
				outs_t = o->prev;
			}
			else {
				o->prev->next = o->next;
				o->next->prev = o->prev;
			}
			free(o);
		}
	}
}

/*
 * flow_control() blocks until the number of outstanding messages and/or
 * bytes on this socket is below a compile time threshold. Signals must 
 * be enabled during this call (rm_lock must not be held).
 */
void flow_control(int s, int len)
{
	int		full_bytes = FALSE;
	int		full_msgs = FALSE;
	fd_set		socks;

	rm_lock();
	check_queues(-1);
	rm_unlock();

	if (bytes_on_sock[s] >= RM_OUT_BYTES_HIGH)
		full_bytes = TRUE;
	if (msgs_on_sock[s] >= RM_OUT_HIGH) 
		full_msgs = TRUE;

	if (full_bytes || full_msgs) {
		while ((full_bytes && (bytes_on_sock[s] > RM_OUT_BYTES_LOW)) ||
		       (full_msgs && (msgs_on_sock[s] > RM_OUT_LOW))) {
			FD_ZERO(&socks);
			FD_SET(s, &socks);
			FD_SET(s+3, &socks);
			select(s+1, &socks, NULL, NULL, NULL);
			FD_CLR(s+3, &socks);
			rm_lock();
			check_queues(-1);
			rm_unlock();
		}
	}

	rm_lock();
	msgs_on_sock[s]++;
	bytes_on_sock[s] += len;
	rm_unlock();
}


/*
 * This function enables asynchronous acking for this connection.
 * A synchronous ack will go out for this connection at the next timer
 * signal, unless a regular message is sent on this connection first.
 * Signals must be disabled (rm_lock held) during this call.
 */
void async_ack_enable(int s, connection_info *c, u_long ip, int port)
{
	if (c->ack_index == -1) {
		if (curr_ack_index < RM_MAX_DESCRIPTORS) {
			c->ack_index = curr_ack_index;
			c->ip = ip;
			c->port = port;
			c->s = s;
			to_ack[curr_ack_index++] = c;
		}
	}
	c->do_ack = TRUE;
}


/*
 * Add a message to our in-order message list. Signals must be off during
 * this call. We'll also check if by adding this message, we can complete
 * a larger sequence with messages in the out-of-order table.
 */
void add_msg(int s, connection_info *c, rm_msg *m, struct sockaddr_in from)
{
	mbuf	*b;
	int	id;
	hash_table *h;
	sockbuf	*sb = &sock_queues[s];

	b = alloc(sizeof(mbuf));
	b->m = m;
	b->from = from;
	b->next = NULL;

	if (sb->head == NULL) {
		sb->head = b;
		sb->tail = b;
	}
	else {
		sb->tail->next = b;
		sb->tail = b;
	}

	/*
	 * Now that we've added the above message to the in-order list,
	 * can we add any more messages from the out-of-order table to
	 * the in-order list?
	 */
	h = c->oo_table;
	id = m->id + 1;
	for (b = hash_find(h, id); b != NULL; id++, b = hash_find(h, id)) {
		b->next = NULL;
		if (sb->head == NULL) {
			sb->head = b;
			sb->tail = b;
		}
		else {
			sb->tail->next = b;
			sb->tail = b;
		}
		hash_delete_entry(h, id);
		c->received++;
		bytes_buffered[s] += b->m->len;
		/*
		 * EOF messages get acked only by the user during
		 * the socket closing protocol.
		 */
		if (b->m->type == RM_EOF)
			c->do_ack = FALSE;
		else {
			u_long	ip = b->from.sin_addr.s_addr;
			int	port = ntohs(b->from.sin_port);

			/*
			 * If necessary, send back an
			 * ack for flow control purposes.
			 */
			c->mcount++;
			if (c->mcount >= RM_ACK_EVERY &&
			    bytes_buffered[s] < RM_BUF_MAX) {
				send_ack(s, c, ip, port, FALSE);
				c->mcount = 0;
			}
		}
	}
}


/*
 * Add a new, but out-of-order message to our out-of-order hash table. 
 * Signals must be off during this call. 
 */
void add_oo_msg(connection_info *c, rm_msg *m, struct sockaddr_in from)
{
	mbuf	*b;

	/*
	 * Check that we haven't already received this oo message
	 */
	b = hash_find(c->oo_table, m->id);
	if (b != NULL) {
		free(m);
		return;
	}

	b = alloc(sizeof(mbuf));
	b->m = m;
	b->from = from;
	b->next = NULL;
	hash_insert(c->oo_table, m->id, b);
}


int send_ack(int s, connection_info *c, u_long ip, int port, int flag)
{
	rm_msg		m;

	/*
	 * Manually build a singleton ack.
	 */
	m.id 	= 0;
	m.type 	= RM_ACK;
	m.len 	= 0;
	m.rcvd 	= c->received;

	m.retrans = flag;

	/*
	 * Send it unreliably
	 */
	return udp_send(s, &m, RM_HEADER_SIZE, 0, ip, port);
}

int send_nack(int s, connection_info *c, u_long ip, int port, u_long id)
{
	rm_msg		m;

	/*
	 * Manually build a singleton ack.
	 */
	m.id 	= id;
	m.type 	= RM_NACK;
	m.len 	= 0;
	m.rcvd 	= c->received;

	/*
	 * Send it unreliably
	 */
	return udp_send(s, &m, RM_HEADER_SIZE, 0, ip, port);
}

/*
 * Retransmit outstanding messages.
 *
 * If who is RM_RETRANS_MSG, the msg 'id' destined for 'ip-port' is resent.
 * If who is RM_RETRANS_HOST, all msgs destined for 'ip-port' are resent.
 * If who is RM_RETRANS_ALL, all outstanding msgs whose tick counters are 
 * above or equal to RM_SIGS_PER_RETRANS are resent.
 */
int retransmit(int who, u_long ip, int port, u_int id)
{
	int		len, res;
	outstanding_msg	*o;

	for (o = outs_h; o != NULL; o = o->next) {
		if (who == RM_RETRANS_MSG && !(o->ip == ip && o->port == port 
		    && o->m->id == id))
			continue;
		if (who == RM_RETRANS_HOST && !(o->ip == ip && o->port == port))
			continue;
		if (who == RM_RETRANS_ALL && (o->ticks < RM_SIGS_PER_RETRANS))
			continue;
		o->m->retrans = TRUE;
		o->ticks = 0;
		len = o->m->len + RM_HEADER_SIZE;

		res = udp_send(o->sock, o->m, len, o->flags, o->ip, o->port);
		if (res < 0) perror("udp_send retrans");

		sock_retrans[o->sock]++;
	}
}

/*
 * Increment all outstanding messages' tick counters.
 */
int retrans_tick(void)
{
	outstanding_msg	*o;

	for (o = outs_h; o != NULL; o = o->next) {
		o->ticks++;
	}
}


/*
 * Check the input queue for socket 's' or all input queues if s is -1.
 * Inspect the heads of each queue for incoming acks and retransmissions.
 * All messages found in the queues are read and handled appropriately. 
 * When this function finishes, either the requested socket queue is 
 * empty, or all socket queues are empty (if sock == -1). Returns 0 on
 * success and -1 on failure. This function should be invoked with signals
 * disabled.
 */
int check_queues(int sock)
{
	int		done, i, s, len, res, port;
	int		fds[RM_MAX_DESCRIPTORS], nfds;
	u_long		ip;
	rm_msg		m;
	fd_set		socks;
	connection_info	*c;
	struct timeval 	poll = {0, 0};
	struct sockaddr_in f;

	int 		maxfd;

	if (sock == -1) {
		for (i=0; i < num_desc; i++)
			fds[i] = descriptors[i];
		nfds = num_desc;
	}
	else {
		fds[0] = sock;
		nfds = 1;
	}

	/*
	 * We will handle all acks and consume retransmissions. We have 
	 * to use select to make sure that we don't block while peeking 
	 * the queues.
	 */
	for (;;) {

		FD_ZERO(&socks);
		for (i=0, maxfd=0; i < nfds; i++) {
			FD_SET(fds[i], &socks);
			if(fds[i]>maxfd)
				maxfd=fds[i];
		}

		FD_SET(maxfd+3, &socks);
		res = select(maxfd+1, &socks, NULL, NULL, &poll);
		FD_CLR(maxfd+3, &socks);

		if (res <= 0) {
			return res;
		}
		for (i=0; i < nfds; i++) {
			if (!FD_ISSET(fds[i], &socks))
				continue;

			/* 
			 * There is stuff to read on this socket, so
			 * peek the first message.
			 */
			s = fds[i];
			len = RM_HEADER_SIZE;
			res = udp_recv(s, &m, len, MSG_PEEK, &f);
			if (res < 0) {
				return -1;
			}

			/*
			trace[trace_i].id = m.id;
			trace[trace_i].retrans = m.retrans;
			trace_i++;
			*/

			ip = f.sin_addr.s_addr;
			port = ntohs(f.sin_port);
			c = get_conn_info(s, ip, port);
			if (m.type == RM_ACK) {
				/* handle this singleton ack and consume it */
				process_acks(ip, port, m.rcvd);
				res = udp_recv(s, &m, len, 0, &f);
				if (res < 0) perror("udp_recv 1");
			}
			else if (m.type == RM_NACK) {
				process_acks(ip, port, m.rcvd);
				retransmit(RM_RETRANS_MSG, ip, port, m.id);
				res = udp_recv(s, &m, len, 0, &f);
				if (res < 0) perror("udp_recv 1");
			}
			else if (!in_wind(m.id, c->received)) { 
				rm_msg	*new;

				/*
				 * A new message. Alloc space for it, 
				 * read in the whole thing, and queue it.
				 */
				len = m.len + RM_HEADER_SIZE;
				new = alloc(len);
				res = udp_recv(s, new, len, 0, &f);
				if (res < 0) perror("udp_recv 2");
				if (new->id != m.id) /* sanity check */
					perror("check_queues: bad msg\n");
				process_acks(ip, port, m.rcvd);
				if (m.id == c->received + 1) {
					/* 
					 * Correct next message. Put it straight
					 * into our in-order msg buffer, and
					 * enable an ack.
					 */
					add_msg(s, c, new, f);
					c->received++;
					async_ack_enable(s, c, ip, port);
					bytes_buffered[s] += len;
					/*
					 * EOFs get acked by the user during
					 * the socket close protocol only.
					 */
					if (m.type == RM_EOF)
						c->do_ack = FALSE;
					else {
						/*
						 * If necessary, send back an
						 * ack for flow control 
						 * purposes. We will not send
						 * back an ack if we've already
						 * maxed out our local buffers.
						 */
						c->mcount++;
						if (c->mcount >= RM_ACK_EVERY &&
						    bytes_buffered[s] < RM_BUF_MAX) {
							send_ack(s,c,ip,port,FALSE);
							c->mcount = 0;
						}
					}
				}
				else {
					u_int	x, y, j;

					/*
					 * We got this msg out of order. Add 
					 * to our out-of-order table. Send 
					 * back a nack for each seemingly
					 * dropped msg. We'll have to do 
					 * things carefully so this algorithm
					 * works for wrapped ids...
					 */
					add_oo_msg(c, new, f);
					x = c->received + 1;
					y = c->nack_seq + 1;
					j = (is_ahead(x, y) ? x : y);
					for (; is_ahead(m.id, j); j++) {
						send_nack(s, c, ip, port, j);
					}
					c->nack_seq = m.id;
				}
			}
			else {
				/* 
				 * This message is a retransmission, so just
				 * consume it. In case the ack for this message
				 * was lost, we'll turn on async acking for 
				 * this msg. 
				 */
				async_ack_enable(c->s, c, ip, port);
				res = udp_recv(s, &m, len, 0, &f);
				if (res < 0) perror("udp_recv 3");
			}
		}
	}
}


/*
 * This is our timer signal handler. All the asynchronous stuff is done here.
 * We send out singleton acks on connections where an ack has not recently
 * gone out (piggybacked on another message for example). We also check
 * for incoming acks at the heads of our socket buffers. Acks down in the
 * buffer behind a new message will not be caught until the new
 * message is consumed by the application. Finally, we retransmit any
 * recently sent messages that have not been acked. Signals must be off 
 * during this call. Signal delivery should take care of that.
 */
void handle_timer(int sig)
{
	int		i, res, len;
	rm_msg		ack;
	connection_info	*c;
	outstanding_msg	*o;
	static int 	sig_count = 0;

	/*
	 * We set this flag in case we're interrupting a blocked call
	 * to rm_select().
	 */
	rm_signal_flag = TRUE;

	/*
	 * Send explicit acks for messages whose acks have
	 * not yet been piggybacked.
	 */
	for (i=0; i < curr_ack_index; i++) {
		c = to_ack[i];
		/* Skip stuff that has already been acked */
		if (!c->do_ack) 
			continue;

		res = send_ack(c->s, c, c->ip, c->port, TRUE);
		if (res < 0) perror("timer send_ack");
	}
	/* Nullify ack list */
	for (i=0; i < curr_ack_index; i++) {
		to_ack[i]->ack_index = -1;
		to_ack[i] = NULL;
	}
	curr_ack_index = 0;

	/*
	 * Read all pending messages, and handle them. New messages will
	 * be added to our internal buffers.
	 */
	check_queues(-1);

	retrans_tick();
	retransmit(RM_RETRANS_ALL, 0, 0, 0);
}


/*
 * Merge descriptors up to n in file descriptor sets s1 and s2, and store
 * the result in dst. Merging computes the logical 'or' of the bits in 
 * the sets.
 */
void merge_sets(int n, fd_set* s1, fd_set* s2, fd_set* dst)
{
	int	i;

	for (i=0; i < n; i++) {
		if (FD_ISSET(i, s1) || FD_ISSET(i, s2))
			FD_SET(i, dst);
	}
}


/*
 * Subtract time t2 from time t1 and store the result in dst. If t2 is
 * bigger than t1, store 0 in dst.
 */
void sub_time(struct timeval* t1, struct timeval* t2, struct timeval* dst)
{
	struct timeval	x, y;

	x = *t1;
	y = *t2;

	/* check that operands can be safely subtracted */
	if (y.tv_sec>x.tv_sec || (y.tv_sec==x.tv_sec && y.tv_usec>x.tv_usec)) {
		dst->tv_sec = 0;
		dst->tv_usec = 0;
		return;
	}

	/* compute x - y */
	if (y.tv_usec > x.tv_usec) {
		/* borrow from seconds, give to usecs */
		x.tv_sec--;
		x.tv_usec += 1000000;
	}
	dst->tv_sec = x.tv_sec - y.tv_sec;
	dst->tv_usec = x.tv_usec - y.tv_usec;
}


/**
 ** Public Functions
 **/

/*
 * rm_init()
 *
 *	Set up rmessages.
 */
void rm_init(void)
{
	int	i, res;
	struct sigaction sa = {handle_timer, 0, SA_RESTART};

	seqno_table = alloc(sizeof(hash_table));
	hash_init(seqno_table, HASH_STRING_KEYS);
	connections = alloc(sizeof(hash_table));
	hash_init(connections, HASH_STRING_KEYS);
	addr_table = alloc(sizeof(hash_table));
	hash_init(addr_table, HASH_ONE_WORD_KEYS);

	outs_h = NULL;
	outs_t = NULL;

	for (i=0; i < RM_MAX_DESCRIPTORS; i++) {
		to_ack[i] = NULL;
		local_ports[i] = 0;
		msgs_on_sock[i] = 0;
		bytes_on_sock[i] = 0;
		sock_retrans[i] = 0;
		sock_queues[i].head = NULL;
		sock_queues[i].tail = NULL;
		is_rm_sock[i] = FALSE;
		bytes_buffered[i] = 0;
	}
	curr_ack_index = 0;

	/* 
	 * Set up timer to signal us every RM_TIMER_PERIOD ms. We do this
	 * one of two ways. We can fork off another process to send
	 * us an RM_TIMER_SIG signal every so often. Doing so lets us get
	 * timer signals without using up an itimer or SIGALRM which the
	 * user might need. This solution does not work under DC which
	 * does not currently support fork(). So for DC applications we
	 * will send a message to a special daemon, informing it that we'd
	 * like to be signaled periodically.
	 */
	sigaction(RM_TIMER_SIG, &sa, NULL);
	rm_lock();

#ifdef BUILD_FOR_DC
	{
	int	daemon_sock, pid, tries;
	u_long	ip;
	fd_set	f;
	struct timeval tout = {RM_DAEMON_TIMEOUT, 0};

	daemon_sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (daemon_sock < 0) {
		perror("rm_init: unable to open daemon socket");
		rm_unlock();
		return;
	}

	ip = get_address("127.0.0.1");
	if (ip == 0) {
		perror("rm_init: unable to form localhost IP");
		rm_unlock();
		return;
	}

	pid = getpid();

	for (tries=0; tries < RM_DAEMON_ATTEMPTS; tries++) {

		res = udp_send(daemon_sock, &pid, sizeof(pid), 0, 
			       ip, RM_DAEMON_PORT);
		if (res < 0) continue;

		/* wait up to RM_DAEMON_TIMEOUT secs for a reply from daemon */
		FD_ZERO(&f);
		FD_SET(daemon_sock, &f);
		FD_SET(daemon_sock+3, &f);
		res = select(daemon_sock+1, &f, NULL, NULL, &tout);
		FD_CLR(daemon_sock+3, &f);

		if (res > 0) {
			/* 
			 * Socket is ready. We don't even have to read the
			 * reply to know that it's an ack from the daemon.
			 */
			close(daemon_sock);
			rm_unlock();
			return;
		}

	}

	if (res < 0)
		perror("rm_init: unable to contact rmessage daemon");
	else
		fprintf(stderr, "rm_init: rmessage daemon not responding\n");

	}
#else

	if (fork() == 0) {
		int	parent, res;
		struct timeval duration = {0, RM_TIMER_PERIOD*1000};

		parent = getppid();
		for (;;) {
			res = select(0, NULL, NULL, NULL, &duration);
			if (res != 0) {
				fprintf(stderr, "timer child select failed\n");
				exit(1);
			}
			else {
				res = kill(parent, RM_TIMER_SIG);
				if (res < 0) {
					exit(0);
				}
			}
		}
		fprintf(stderr, "Timer child exiting?!?\n");
		exit(1);
	}

#endif

	rm_unlock();
}


/*
 * rm_socket()
 *
 *	Return an open socket descriptor for use in later messaging calls,
 *	or -1 on failure.
 */
int rm_socket(void)
{
	int	d, res;

	if (num_desc == RM_MAX_DESCRIPTORS) {
		errno = EMFILE;
		return -1;
	}

	rm_lock();

	d = socket(AF_INET, SOCK_DGRAM, 0);
	if (d >= 0) {
		int	bufsize = RM_SOCKRBUF_SIZE;
		res = setsockopt(d,SOL_SOCKET,SO_RCVBUF,&bufsize,sizeof(int));
		if (res < 0) {
			close(d);
			rm_unlock();
			return -1;
		}
		descriptors[num_desc++] = d;
		is_rm_sock[d] = TRUE;
	}

	rm_unlock();

	return d;
}


/*
 * rm_bind()
 *
 *	Bind a socket to a specific port. If the provided port number
 *	is zero, the system will choose an available port for the socket.
 *	Returns 0 on success or -1 on failure. 
 */
int rm_bind(int sock, int port)
{
	struct sockaddr_in addr;

	bzero(&addr, sizeof(addr));
	addr.sin_family 	= AF_INET;
	addr.sin_addr.s_addr 	= htonl(INADDR_ANY);
	addr.sin_port 		= htons(port);

	return bind(sock, (struct sockaddr*) &addr, sizeof(addr));
}


int rm_connect_ip(int sock, u_long ip, int port)
{
	int	res;
	struct sockaddr_in addr, *a;

	a = hash_find(addr_table, (char*)sock);
	if (a == NULL) {
		a = alloc(sizeof(struct sockaddr_in));
		hash_insert(addr_table, (char*)sock, a);
	}

	bzero(a, sizeof(*a));
	a->sin_family 		= AF_INET;
	a->sin_port 		= htons(port);
	a->sin_addr.s_addr 	= ip;

	return 0;
}


/*
 * rm_connect()
 *
 *	Connect a UDP socket to a named host. By connecting a connectionless
 *	socket in this manner, the user is really just specifying a 
 *	destination address so that later calls to send() and write()
 *	(functions that do not take a destination address parameter) know
 *	where to send their data. It returns 0 on success and -1 on failure.
 */
int rm_connect(int sock, char *host, int port)
{
	unsigned long ip;

	ip = get_address(host);
	if (!ip) 
		return -1;

	return rm_connect_ip(sock, ip, port);
}


/*
 * rm_fill_sockaddr()
 *
 *	This routine is provided as a convenience to the user. It takes
 *	a host address (a string consisting of a hostname or ip address 
 *	in dotted decimal format), and port and fills in a sockaddr_in
 *	struct appropriately. It returns -1 if the hostname is invalid
 *	and 0 otherwise.
 */
int rm_fill_sockaddr(struct sockaddr_in *a, char* host, int port)
{
	u_long	ip;

	ip = get_address(host);
	if (!ip)
		return -1;

	a->sin_family		= AF_INET;
	a->sin_port 		= htons(port);
	a->sin_addr.s_addr 	= ip;
	return 0;
}


/*
 * rm_sendto()
 *
 *	Sends a reliable datagram on a connected socket. Datagrams will be
 *	delivered on the receiving side in the same order in which they
 *	were sent, and packet losses will be invisible to the user. Returns
 *	0 on success and -1 on failure. 
 */
int rm_sendto(int sock, void *msg, int len, int flags, struct sockaddr_in *to)
{
	int	port, res;
	u_long	ip;
	rm_msg	*m;

	/* 
	 * If necessary, wait for consumer to catch up before 
	 * enqueueing new msg 
	 */
	flow_control(sock, len + RM_HEADER_SIZE); 

	rm_lock();

	port = ntohs(to->sin_port);
	ip = to->sin_addr.s_addr;
	m = build_msg(sock, msg, len, ip, port, flags);
	add_msg_to_tables(m, sock, ip, port, flags);
	len += RM_HEADER_SIZE;

	rm_unlock();

	return udp_send(sock, m, len, flags, ip, port);
}


/*
 * rm_recvp()
 *
 *	Receives a reliable datagram from a socket. These datagrams are
 *	delivered in the same order in which they were sent. Returns a
 *	pointer to the message (an rm_msg struct) on success, and NULL
 *	on failure.
 */
rm_msg* rm_recvp(int s, int flags, struct sockaddr_in *from)
{
	sockbuf		*sb = &sock_queues[s];
	fd_set		socks;
	rm_msg		*m;
	mbuf		*b;

	rm_lock();

	/*
	 * Read all available messages into our internal buffers
	 */
	check_queues(s);

	/*
	 * Check if there is a message to return, if not, block
	 */
	while (sb->head == NULL) {
		FD_ZERO(&socks);
		FD_SET(s, &socks);
		FD_SET(s+3, &socks);
		rm_unlock();
		select(s+1, &socks, NULL, NULL, NULL);
		rm_lock();
		FD_CLR(s+3, &socks);
		check_queues(s);
	}

	/*
	 * Take first item off in-order list
	 */
	b = sb->head;
	if (!(flags & MSG_PEEK)) {
		sb->head = b->next;
		if (b->next == NULL)
			sb->tail = NULL;
	}

	if (from != NULL)
		*from = b->from;
	m = b->m;

	if (!(flags & MSG_PEEK)) {
		if (m->type == RM_EOF) {
			u_long	ip = b->from.sin_addr.s_addr;
			int	port = ntohs(b->from.sin_port);
			connection_info *c = get_conn_info(s, ip, port);
			c->closing = TRUE;
		}
		free(b);
		bytes_buffered[s] -= m->len;
	}

	rm_unlock();
	return m;
}


/*
 * rm_recvfrom()
 *
 *	Receives a reliable datagram from a socket. These datagrams are
 *	delivered in the same order in which they were sent. Returns the
 *	number of bytes received on success and -1 on failure.
 */
int rm_recvfrom(int s, void *b, int len, int flags, struct sockaddr_in *from)
{
	int	tocopy;
	rm_msg	*m;


	m = rm_recvp(s, flags, from);
	if (m == NULL) return -1;

	tocopy = (m->len > len ? len : m->len);
	bcopy(m->data, b, tocopy);

	/*
	 * Block signals around free() so we don't reenter malloc routines
	 * if a signal comes in while the free is being performed.
	 */
	rm_lock();

	if (!(flags & MSG_PEEK))
		free(m);

	rm_unlock();

	return tocopy;
}


/*
 * rm_recv()
 *
 *	Receives a reliable datagram from a socket. These datagrams are
 *	delivered in the same order in which they were sent. Returns the
 *	number of bytes received on success and -1 on failure.
 */
int rm_recv(int s, void *b, int len, int flags)
{
	return rm_recvfrom(s, b, len, flags, NULL);
}


/*
 * rm_send()
 *
 *	Sends a reliable datagram on a connected socket. Datagrams will be
 *	delivered on the receiving side in the same order in which they
 *	were sent, and packet losses will be invisible to the user. Returns
 *	0 on success and -1 on failure. 
 */
int rm_send(int sock, void *msg, int len, int flags)
{
	struct sockaddr_in *to;

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) {
		errno = ENOTCONN;
		return -1;
	}

	return rm_sendto(sock, msg, len, flags, to);
}


/*
 * rm_select()
 *
 *	Perform all the usual duties of select() on both rmessage
 *	socket descriptors, and other file descriptors. Invoked and
 *	used exactly like regular select(). Rmessage descriptors
 *	do not have exception conditions kept for them, nor is it 
 *	possible to do non-blocking I/O on them. Therefore, the writefds
 *	and exceptfds sets are useless for Rmessage sockets.
 */
int rm_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
              struct timeval *timeout)
{
	int		i, fd, count, res;
	fd_set		outfds, r, w, e;
	sockbuf		*sb;
	struct timeval	start, curr, diff, poll={0, 0}, *tout, foo;

	if (nfds > RM_MAX_DESCRIPTORS)
		nfds = RM_MAX_DESCRIPTORS;

	if (timeout == NULL)
		tout = NULL;
	else {
		foo = *timeout;
		tout = &foo;
	}

	/*
	 * Note the time so we can tell how long we've actually been
	 * blocked if we're awakened early by a signal, or meta-message.
	 */
	if (gettimeofday(&start, NULL) < 0) {
		fprintf(stderr, "rm_select: gettimeofday() 1 failed\n");
		return -1;
	}

	if (readfds != NULL) r = *readfds; 
	if (writefds != NULL) w = *writefds; 
	if (exceptfds != NULL) e = *exceptfds;

	for (;;) {

		rm_lock();

		/*
		 * Process all pending rmessages
		 */
		check_queues(-1);

		/*
		 * Start to build up read fd set. First check named rm socks
		 * for currently queued messages.
		 */
		if (readfds != NULL) {
			count = 0;
			FD_ZERO(&outfds);
			for (i = 0; i < num_desc; i++) {
				fd = descriptors[i];
				if (fd < nfds && FD_ISSET(fd, readfds)) {
					sb = &sock_queues[fd];
					if (sb->head != NULL) {
						FD_SET(fd, &outfds);
						count++;
					}
				}
			}
		}

		/*
		 * Now quickly check all non-rm descriptors for read readiness
		 */
		if (readfds != NULL) {
			int	new_res;
			fd_set	rfds;

			/* build new fd set w/o rm socks in it */
			FD_ZERO(&rfds);
			for (i=0; i < nfds; i++) {
				if (FD_ISSET(i, readfds) && !is_rm_sock[i]) {
					FD_SET(i, &rfds);
				}
			}
			
			FD_SET(nfds+2, &rfds);
			res = select(nfds, &rfds, writefds, exceptfds, &poll);
			FD_CLR(nfds+2, &rfds);

			if (res < 0) {
				perror("rm_select: select 1 failed");
				rm_unlock();
				return -1;
			}
			new_res = res + count;
			if (new_res != 0) {
				FD_ZERO(readfds);
				merge_sets(nfds, &outfds, &rfds, readfds);
				rm_unlock();
				return new_res;
			}
		}

		/*
		 * If we get to this point, then none of the named 
		 * descriptors have anything ready on them. We'll have
		 * to block. We'll reset the fd sets and do a blocking 
		 * select.
		 */
		if (readfds != NULL) *readfds = r; 
		if (writefds != NULL) *writefds = w; 
		if (exceptfds != NULL) *exceptfds = e;
		rm_signal_flag = FALSE;
		rm_unlock();

		if(readfds) FD_SET(nfds+2, readfds);
		res = select(nfds, readfds, writefds, exceptfds, tout);
		if(readfds) FD_CLR(nfds+2, readfds);

		if (res < 0 && !(errno == EINTR && rm_signal_flag)) {
			perror("rm_select: select 2 failed");
			return -1;
		}
		else if (res == 0) 
			return 0;

		/*
		 * Deduct used portion of timeout in case we block again.
		 */
		if (tout == NULL) continue;
		if (gettimeofday(&curr, NULL) < 0) {
			fprintf(stderr, "rm_select: gettimeofday() 2 failed\n");
			return -1;
		}
		sub_time(&curr, &start, &diff);
		*tout = *timeout;
		sub_time(tout, &diff, tout);

	}
}


/*
 * rm_get_mtu()
 *
 *	Return the maximum number of bytes sendable in a single message 
 *	on a particular socket (the "maximum transfer unit" or MTU). The
 *	mtu is based on the sendbuffer size which the user can adjust
 * 	using the setsockopt() call.
 */
int rm_get_mtu(int s)
{
	int	len, res, bufsize, mtu;

	len = sizeof(bufsize);
	res = getsockopt(s, SOL_SOCKET, SO_SNDBUF, &bufsize, &len);
	if (res < 0) 
		return -1;
	else {
		mtu = bufsize - RM_HEADER_SIZE; 
		if (mtu > RM_MAX_MTU)
			return RM_MAX_MTU;
		else
			return mtu;
	}
}


/*
 * rm_total_bytes_ready()
 *
 *	This function returns the number of bytes that are available to be
 * 	returned on the given socket. It consumes packets received out 
 *	of order, as well as singleton ack messages. It returns the number 
 *	of bytes available, or -1 on failure. If no messages are ready to 
 *	be read, this function will return -1 with errno set to ENOBUFS. 
 *	This is to help differentiate no ready data from one or more 0 length
 *	messages ready to be read.
 */
int rm_total_bytes_ready(int s)
{
	int	res = 0;
	sockbuf	*sb = &sock_queues[s];
	mbuf	*b;

	rm_lock();

	check_queues(s);

	if (sb->head == NULL) {
		res = -1;
		errno = ENOBUFS;
	}
	else {
		for (b = sb->head; b != NULL; b = b->next)
			res += b->m->len;
	}
	
	rm_unlock();

	return res;
}

/*
 * rm_bytes_ready()
 *
 *	This function returns the number of bytes that are available to be
 * 	returned on the given socket's first available message, or -1 on 
 *	failure. If no messages are ready to be read, this function will 
 * 	return -1 with errno set to ENOBUFS. This is to help differentiate 
 *	no ready data from one or more 0 length messages ready to be read.
 */
int rm_bytes_ready(int s)
{
	int	res = 0;
	sockbuf	*sb = &sock_queues[s];
	mbuf	*b;

	rm_lock();

	check_queues(s);

	if (sb->head == NULL) {
		res = -1;
		errno = ENOBUFS;
	}
	else {
		res = sb->head->m->len;
	}
	
	rm_unlock();

	return res;
}


/*
 * Returns the number of messages waiting on socket 's'.
 */
int rm_msgs_ready(int s)
{
	int	res = 0;
	sockbuf	*sb = &sock_queues[s];
	mbuf	*b;

	rm_lock();

	check_queues(s);

	for (b = sb->head; b != NULL; b = b->next)
		res++;
	
	rm_unlock();

	return res;
}


/*
 * rm_drain()
 *
 *	Waits for all outgoing messages on 'sock' to be received and
 *	acknowledged by their recipients.
 */
void rm_drain(int sock)
{
	int		msgs, i;
	outstanding_msg *o;
	struct timeval	t = {0, 100000}; /* 100 ms */

	for (;;) {
		msgs = 0;
		rm_lock();
		for (o = outs_h; o != NULL; o = o->next) {
			if (o->sock == sock) msgs++;
		}
		rm_unlock();
		if (msgs == 0) 
			break;
		else {
			/*
			 * There is still stuff outstanding, so wait for a
			 * while then check again. Timer signals will likely
			 * wake this call early...
			 */
			select(0, NULL, NULL, NULL, &t);
		}
	}
}


/*
 * rm_close()
 *
 * 	Close a socket. This function cleans up internal data structures
 *	and attempts to ensure that all the messages sent on this socket
 *	have been received.
 */
int rm_close(int sock)
{
	int		msgs, i;
	outstanding_msg	*o;
	struct timeval	t = {0, 100000};
	sockbuf		*sb = &sock_queues[sock];
	mbuf		*b, *n;

	rm_lock();

	/*
	 * We don't have enough information here to remove the entries
	 * in the "connections" table and the "seqno" table, each of which
	 * are indexed by the full (sock, ip, port) of the peer. We'll
	 * assume here that if a new connection if opened to the same peer,
	 * it will at least have a new port number (possibly also a new 
	 * socket number), so that new clean entries in those two tables
	 * will be automatically created. 
	 */

	/*
	 * Remove address info from connection address table.
	 */
	hash_delete_entry(addr_table, (char*)sock);

	/* print retrans info */
	fprintf(stderr, "rm_close: sock %d had %d retransmissions\n", sock,
		sock_retrans[sock]);

	/*
	 * Reset per socket info.
	 */
	local_ports[sock] = 0;
	msgs_on_sock[sock] = 0;
	bytes_on_sock[sock] = 0;
	sock_retrans[sock] = 0;
	is_rm_sock[sock] = FALSE;

	/*
	 * Free undelivered messages on this sock
	 */
	for (b = sb->head; b != NULL; b = n) {
		n = b->next;
		free(b->m);
		free(b);
	}
	sb->head = NULL;
	sb->tail = NULL;

	/*
	 * Remove this socket from our list of open fds.
	 */
	for (i=0; i < num_desc; i++)
		if (descriptors[i] == sock) 
			break;
	if (i == num_desc)
		fprintf(stderr, "warning: socket fd not found in open fds.\n");
	else {
		for (; i < num_desc; i++)
			descriptors[i] = descriptors[i+1];
		num_desc--;
	}

	rm_unlock();

	return close(sock);
}


/*
 * rm_conn_closing()
 *
 *	Returns true if sock's connection has already received an EOF
 *	message, and false otherwise. Returns FALSE if called on a
 *	non-connected socket.
 */
int rm_conn_closing(int sock)
{
	u_long		ip;
	int		port;
	connection_info	*c;
	struct sockaddr_in *to;

	rm_lock();

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) 
		return FALSE;
	
	ip = to->sin_addr.s_addr;
	port = ntohs(to->sin_port);
	c = get_conn_info(sock, ip, port);

	rm_unlock();

	return c->closing;
}


/*
 * rm_ack()
 *
 *	This function sends a current singleton ack on the provided
 *	connected socket.
 */
int rm_ack(int sock)
{
	u_long		ip;
	int		port;
	rm_msg		m;
	connection_info	*c;
	struct sockaddr_in *to;

	rm_lock();

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) {
		errno = ENOTCONN;
		return -1;
	}
	
	ip = to->sin_addr.s_addr;
	port = ntohs(to->sin_port);
	c = get_conn_info(sock, ip, port);

	rm_unlock();

	return send_ack(sock, c, ip, port, FALSE);
}


/*
 * Provides the functionality of the getpeername() syscall for an
 * rmessage socket.
 */
int rm_getpeername(int sock, struct sockaddr_in* name)
{
	u_long		ip;
	int		port;
	rm_msg		m;
	connection_info	*c;
	struct sockaddr_in *to;

	rm_lock();

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) {
		errno = ENOTCONN;
		return -1;
	}

	*name = *to;
	rm_unlock();
	return 0;
}

void print_trace(void) 
{
	int	i;

	fprintf(stderr, "\n\nMessage Trace:\n\n");

	for (i=0; i < trace_i; i++) {
		fprintf(stderr, "%s\t%d %s\n", (i != 0 && trace[i].id != (trace[i-1].id + 1) ? "**" : ""), trace[i].id, (trace[i].retrans ? "RE" : ""));
	}
}


